/* SPDX-License-Identifier: GPL-2.0-only */
/*
 * Copyright (c) 2016-2021, The Linux Foundation. All rights reserved.
 * Copyright (c) 2022-2023, Qualcomm Innovation Center, Inc. All rights reserved.
 */

#ifndef _CAM_REQ_MGR_CORE_H_
#define _CAM_REQ_MGR_CORE_H_

#include <linux/spinlock_types.h>
#include "cam_req_mgr_interface.h"
#include "cam_req_mgr_core_defs.h"
#include "cam_req_mgr_workq.h"
#include "cam_req_mgr_timer.h"

#define CAM_REQ_MGR_MAX_LINKED_DEV     16
#define MAX_REQ_SLOTS                  48
#define MAX_REQ_STATE_MONITOR_NUM      108
#define MAX_DEV_FOR_SPECIAL_OPS        4

#define CAM_REQ_MGR_WATCHDOG_TIMEOUT          1000
#define CAM_REQ_MGR_WATCHDOG_TIMEOUT_DEFAULT  5000
#define CAM_REQ_MGR_WATCHDOG_TIMEOUT_MAX      50000
#define CAM_REQ_MGR_SCHED_REQ_TIMEOUT         1000
#define CAM_REQ_MGR_SIMULATE_SCHED_REQ        30
#define CAM_REQ_MGR_DEFAULT_HDL_VAL           0

#define FORCE_DISABLE_RECOVERY  2
#define FORCE_ENABLE_RECOVERY   1
#define AUTO_RECOVERY           0

#define CRM_WORKQ_NUM_TASKS 60

#define MAX_SYNC_COUNT 65535


/* Default frame rate is 30 */
#define DEFAULT_FRAME_DURATION 33333333

#define INITIAL_IN_SYNC_REQ 5

#define SYNC_LINK_SOF_CNT_MAX_LMT 1

#define MAXIMUM_LINKS_CAPACITY     8

#define MAXIMUM_LINKS_PER_SESSION  4

#define MAXIMUM_RETRY_ATTEMPTS 3

#define VERSION_1  1
#define VERSION_2  2
#define CAM_REQ_MGR_MAX_TRIGGERS   2

#define REQ_MAXIMUM_BUBBLE_TIMES   2

#define CAM_REQ_MGR_HALF_FRAME_DURATION(frame_duration) (frame_duration / 2)

/* Number of words for dumping req state info*/
#define CAM_CRM_DUMP_EVENT_NUM_WORDS  6

/* Maximum length of tag while dumping */
#define CAM_CRM_DUMP_TAG_MAX_LEN 128

/* Number of headers in dumping req state info function*/
#define CAM_CRM_DUMP_EVENT_NUM_HEADERS  2


/**
 * enum crm_workq_task_type
 * @codes: to identify which type of task is present
 */
enum crm_workq_task_type {
	CRM_WORKQ_TASK_GET_DEV_INFO,
	CRM_WORKQ_TASK_SETUP_LINK,
	CRM_WORKQ_TASK_DEV_ADD_REQ,
	CRM_WORKQ_TASK_APPLY_REQ,
	CRM_WORKQ_TASK_NOTIFY_SOF,
	CRM_WORKQ_TASK_NOTIFY_EOF,
	CRM_WORKQ_TASK_NOTIFY_ERR,
	CRM_WORKQ_TASK_NOTIFY_FREEZE,
	CRM_WORKQ_TASK_SCHED_REQ,
	CRM_WORKQ_TASK_FLUSH_REQ,
	CRM_WORKQ_TASK_INVALID,
};

/**
 * struct crm_task_payload
 * @type           : to identify which type of task is present
 * @u              : union of payload of all types of tasks supported
 * @sched_req      : contains info of  incoming reqest from CSL to CRM
 * @flush_info     : contains info of cancelled reqest
 * @dev_req        : contains tracking info of available req id at device
 * @send_req       : contains info of apply settings to be sent to devs in link
 * @notify_trigger : contains notification from IFE to CRM about trigger
 * @notify_err     : contains error info happened while processing request
 * -
 */
struct crm_task_payload {
	enum crm_workq_task_type type;
	union {
		struct cam_req_mgr_sched_request_v2     sched_req;
		struct cam_req_mgr_flush_info           flush_info;
		struct cam_req_mgr_add_request          dev_req;
		struct cam_req_mgr_send_request         send_req;
		struct cam_req_mgr_trigger_notify       notify_trigger;
		struct cam_req_mgr_error_notify         notify_err;
	} u;
};

/**
 * enum crm_req_state
 * State machine for life cycle of request in pd table
 * EMPTY   : indicates req slot is empty
 * PENDING : indicates req slot is waiting for reqs from all devs
 * READY   : indicates req slot is ready to be sent to devs
 * INVALID : indicates req slot is not in valid state
 */
enum crm_req_state {
	CRM_REQ_STATE_EMPTY,
	CRM_REQ_STATE_PENDING,
	CRM_REQ_STATE_READY,
	CRM_REQ_STATE_INVALID,
};

/**
 * enum crm_slot_status
 * State machine for life cycle of request in input queue
 * NO_REQ     : empty slot
 * REQ_ADDED  : new entry in slot
 * REQ_PENDING    : waiting for next trigger to apply
 * REQ_READY      : req has ready
 * REQ_APPLIED    : req is sent to all devices
 * INVALID    : invalid state
 */
enum crm_slot_status {
	CRM_SLOT_STATUS_NO_REQ,
	CRM_SLOT_STATUS_REQ_ADDED,
	CRM_SLOT_STATUS_REQ_PENDING,
	CRM_SLOT_STATUS_REQ_READY,
	CRM_SLOT_STATUS_REQ_APPLIED,
	CRM_SLOT_STATUS_INVALID,
};

/**
 * enum cam_req_mgr_link_state
 * State machine for life cycle of link in crm
 * AVAILABLE  : link available
 * IDLE       : link initialized but not ready yet
 * READY      : link is ready for use
 * ERR        : link has encountered error
 * MAX        : invalid state
 */
enum cam_req_mgr_link_state {
	CAM_CRM_LINK_STATE_AVAILABLE,
	CAM_CRM_LINK_STATE_IDLE,
	CAM_CRM_LINK_STATE_READY,
	CAM_CRM_LINK_STATE_ERR,
	CAM_CRM_LINK_STATE_MAX,
};

/**
 * enum cam_req_mgr_sync_type
 * Sync type for syncing info
 * DELAY_AT_SOF  : inject delay at SOF
 * DELAY_AT_EOF  : inject delay at EOF
 * APPLY_AT_EOF  : apply at EOF
 */
enum cam_req_mgr_sync_type {
	CAM_SYNC_TYPE_DELAY_AT_SOF,
	CAM_SYNC_TYPE_DELAY_AT_EOF,
	CAM_SYNC_TYPE_APPLY_AT_EOF,
};

/**
 * enum cam_req_mgr_req_state
 * Request operation type for request state recording
 * REQ_READY        : all packets for this request is added
 * NOTIFY_TRIGGER   : request is triggered
 * PROCESS_TRIGGER  : request is in trigger processing
 * SEND_REQ         : CRM is sending request to each device
 * NOTIFY_ERR       : request is notified an error
 * PROCESS_ERR      : request is in error processing
 * INVALID          : invalid state
 */
enum cam_req_mgr_req_state {
	CAM_CRM_REQ_READY,
	CAM_CRM_NOTIFY_TRIGGER,
	CAM_CRM_PROCESS_TRIGGER,
	CAM_CRM_SEND_REQ,
	CAM_CRM_NOTIFY_ERR,
	CAM_CRM_PROCESS_ERR,
	CAM_CRM_STATE_INVALID,
};

/**
 * struct cam_req_mgr_traverse_result
 * @req_id        : Req id that is not ready
 * @pd            : pipeline delay
 * @masked_value  : Holds the dev bit for devices not ready
 *                  for the given request
 */
struct cam_req_mgr_traverse_result {
	int64_t  req_id;
	uint32_t pd;
	uint32_t masked_value;
};

/**
 * struct cam_req_mgr_traverse
 * @idx              : slot index
 * @result           : contains which all tables were able to apply successfully
 * @result_data      : holds the result of traverse in case it fails
 * @tbl              : pointer of pipeline delay based request table
 * @apply_data       : pointer which various tables will update during traverse
 * @in_q             : input request queue pointer
 * @validate_only    : Whether to validate only and/or update settings
 * @open_req_cnt     : Count of open requests yet to be serviced in the kernel.
 */
struct cam_req_mgr_traverse {
	int32_t                            idx;
	uint32_t                           result;
	struct cam_req_mgr_traverse_result result_data;
	struct cam_req_mgr_req_tbl        *tbl;
	struct cam_req_mgr_apply          *apply_data;
	struct cam_req_mgr_req_queue      *in_q;
	bool                               validate_only;
	uint32_t                           open_req_cnt;
};

/**
 * struct cam_req_mgr_apply
 * @idx      : corresponding input queue slot index
 * @pd       : pipeline delay of device
 * @req_id   : req id for dev with above pd to process
 * @skip_idx: skip applying settings when this is set.
 */
struct cam_req_mgr_apply {
	int32_t idx;
	int32_t pd;
	int64_t req_id;
	int32_t skip_idx;
};

/**
 * struct crm_tbl_slot_special_ops
 * @num_dev         : Number of devices need to be applied at this trigger point
 * @dev_hdl         : Device handle who requested for special ops
 * @apply_at_eof    : Boolean Identifier for request to be applied at EOF
 * @is_applied      : Flag to identify if request is already applied to device
 *                    in previous frame
 * @skip_isp_apply  : Flag to indicate skip apply req for ISP
 */
struct crm_tbl_slot_special_ops {
	uint32_t num_dev;
	int32_t dev_hdl[MAX_DEV_FOR_SPECIAL_OPS];
	bool apply_at_eof;
	bool is_applied;
	bool skip_isp_apply;
};

/**
 * struct cam_req_mgr_tbl_slot
 * @idx                 : slot index
 * @req_ready_map       : mask tracking which all devices have request ready
 * @req_apply_map       : mask tracking which all devices have request applied
 * @state               : state machine for life cycle of a slot
 * @inject_delay_at_sof : insert extra bubbling for flash type of use cases
 * @inject_delay_at_eof : insert extra bubbling for flash type of use cases
 * @ops                 : special operation for the table slot
 *                        e.g.
 *                        skip_next frame: in case of applying one device
 *                        and skip others
 *                        apply_at_eof: device that needs to apply at EOF
 */
struct cam_req_mgr_tbl_slot {
	int32_t                                idx;
	uint32_t                               req_ready_map;
	uint32_t                               req_apply_map;
	enum crm_req_state                     state;
	uint32_t                               inject_delay_at_sof;
	uint32_t                               inject_delay_at_eof;
	struct  crm_tbl_slot_special_ops       ops;
};

/**
 * struct cam_req_mgr_req_tbl
 * @id            : table indetifier
 * @pd            : pipeline delay of table
 * @dev_count     : num of devices having same pipeline delay
 * @dev_mask      : mask to track which devices are linked
 * @skip_traverse : to indicate how many traverses need to be dropped
 *                  by this table especially in the beginning or bubble recovery
 * @next          : pointer to next pipeline delay request table
 * @pd_delta      : differnce between this table's pipeline delay and next
 * @num_slots     : number of request slots present in the table
 * @slot          : array of slots tracking requests availability at devices
 */
struct cam_req_mgr_req_tbl {
	int32_t                     id;
	int32_t                     pd;
	int32_t                     dev_count;
	int32_t                     dev_mask;
	int32_t                     skip_traverse;
	struct cam_req_mgr_req_tbl *next;
	int32_t                     pd_delta;
	int32_t                     num_slots;
	struct cam_req_mgr_tbl_slot slot[MAX_REQ_SLOTS];
};

/**
 * struct cam_req_mgr_slot
 * - Internal Book keeping
 * @idx                : slot index
 * @skip_idx           : if req id in this slot needs to be skipped/not applied
 * @status             : state machine for life cycle of a slot
 * - members updated due to external events
 * @recover            : if user enabled recovery for this request.
 * @req_id             : mask tracking which all devices have request ready
 * @sync_mode          : Sync mode in which req id in this slot has to applied
 * @additional_timeout : Adjusted watchdog timeout value associated with
 * this request
 * @recovery_counter   : Internal recovery counter
 * @num_sync_links     : Num of sync links
 * @sync_link_hdls     : Array of sync link handles
 * @bubble_times       : times of bubbles the req happended
 * @internal_recovered : indicate if internal recover is already done for request
 * of this slot
 */
struct cam_req_mgr_slot {
	int32_t               idx;
	int32_t               skip_idx;
	enum crm_slot_status  status;
	int32_t               recover;
	int64_t               req_id;
	int32_t               sync_mode;
	int32_t               additional_timeout;
	int32_t               recovery_counter;
	int32_t               num_sync_links;
	int32_t               sync_link_hdls[MAXIMUM_LINKS_PER_SESSION - 1];
	uint32_t              bubble_times;
	ktime_t               ts;
	ktime_t               apply_ts;
	bool                  internal_recovered;
};

/**
 * struct cam_req_mgr_req_queue
 * @num_slots   : max num of input queue slots
 * @slot        : request slot holding incoming request id and bubble info.
 * @rd_idx      : indicates slot index currently in process.
 * @wr_idx      : indicates slot index to hold new upcoming req.
 * @last_applied_idx : indicates slot index last applied successfully.
 */
struct cam_req_mgr_req_queue {
	int32_t                     num_slots;
	struct cam_req_mgr_slot     slot[MAX_REQ_SLOTS];
	int32_t                     rd_idx;
	int32_t                     wr_idx;
	int32_t                     last_applied_idx;
};

/**
 * struct cam_req_mgr_req_state_monitor
 * @req_state    : operation type
 * @req_id       : request id
 * @dev_hdl      : to track which device is sending request
 * @frame_id     : frame id
 * @time_stamp   : the time stamp of the state
 * @name         : device_name
 */
struct cam_req_mgr_state_monitor {
	enum cam_req_mgr_req_state  req_state;
	int64_t                     req_id;
	int32_t                     dev_hdl;
	int64_t                     frame_id;
	struct timespec64           time_stamp;
	char                        name[256];
};

/**
 * struct cam_req_mgr_req_data
 * @in_q                  : Poiner to Input request queue
 * @l_tbl                 : unique pd request tables.
 * @num_tbl               : how many unique pd value devices are present
 * @apply_data            : Holds information about request id for a request
 * @prev_apply_data       : Holds information about request id for a previous
 *                          applied request
 * @state_monitor         : used to track the request state in CRM
 * @next_state_idx        : indicates index to hold new request state
 * @lock                  : mutex lock protecting request data ops.
 * @monitor_slock         : spin lock protecting state mointor variables
 * @reset_link_spin_lock  : spin lock protecting link data ops when reset link
 */
struct cam_req_mgr_req_data {
	struct cam_req_mgr_req_queue *in_q;
	struct cam_req_mgr_req_tbl   *l_tbl;
	int32_t                       num_tbl;
	struct cam_req_mgr_apply      apply_data[CAM_PIPELINE_DELAY_MAX];
	struct cam_req_mgr_apply      prev_apply_data[CAM_PIPELINE_DELAY_MAX];
	struct cam_req_mgr_state_monitor state_monitor[
		MAX_REQ_STATE_MONITOR_NUM];
	int32_t                       next_state_idx;
	struct mutex                  lock;
	spinlock_t                    monitor_slock;
	spinlock_t                    reset_link_spin_lock;
};

/**
 * struct cam_req_mgr_connected_device
 * - Device Properties
 * @dev_hdl   : device handle
 * @dev_bit   : unique bit assigned to device in link
 * - Device characteristics
 * @pd_tbl    : tracks latest available req id at this device
 * @dev_info  : holds dev characteristics such as pipeline delay, dev name
 * @ops       : holds func pointer to call methods on this device
 * @parent    : pvt data - like link which this dev hdl belongs to
 * @is_active : indicate whether device is active in auto shdr usecase
 */
struct cam_req_mgr_connected_device {
	int32_t                         dev_hdl;
	int64_t                         dev_bit;
	struct cam_req_mgr_req_tbl     *pd_tbl;
	struct cam_req_mgr_device_info  dev_info;
	struct cam_req_mgr_kmd_ops     *ops;
	void                           *parent;
	bool                            is_active;
};

/**
 * struct cam_req_mgr_core_link
 * -  Link Properties
 * @link_hdl             : Link identifier
 * @num_devs             : num of connected devices to this link
 * @max_delay            : Max of pipeline delay of all connected devs
 * @min_delay            : Min of pipeline delay of all connected devs
 * @max_mswitch_delay    : Max of modeswitch delay of all connected devs
 * @min_mswitch_delay    : Min of modeswitch delay of all connected devs
 * @workq                : Pointer to handle workq related jobs
 * @pd_mask              : each set bit indicates the device with pd equal to
 *                          bit position is available.
 * - List of connected devices
 * @l_dev                : List of connected devices to this link
 * - Request handling data struct
 * @req                  : req data holder.
 * - Timer
 * @watchdog             : watchdog timer to recover from sof freeze
 * - Link private data
 * @workq_comp           : conditional variable to block user thread for workq
 *                          to finish schedule request processing
 * @state                : link state machine
 * @parent               : pvt data - link's parent is session
 * @lock                 : mutex lock to guard link data operations
 * @link_state_spin_lock : spin lock to protect link state variable
 * @sync_link            : array of pointer to the sync link for synchronization
 * @num_sync_links       : num of links sync associated with this link
 * @sync_link_sof_skip   : flag determines if a pkt is not available for a given
 *                         frame in a particular link skip corresponding
 *                         frame in sync link as well.
 * @open_req_cnt         : Counter to keep track of open requests that are yet
 *                         to be serviced in the kernel.
 * @last_flush_id        : Last request to flush
 * @is_used              : 1 if link is in use else 0
 * @is_master            : Based on pd among links, the link with the highest pd
 *                         is assigned as master
 * @initial_skip         : Flag to determine if slave has started streaming in
 *                         master-slave sync
 * @in_msync_mode        : Flag to determine if a link is in master-slave mode
 * @initial_sync_req     : The initial req which is required to sync with the
 *                         other link
 * @retry_cnt            : Counter that tracks number of attempts to apply
 *                         the same req
 * @is_shutdown          : Flag to indicate if link needs to be disconnected
 *                         as part of shutdown.
 * @sof_timestamp_value  : SOF timestamp value
 * @prev_sof_timestamp   : Previous SOF timestamp value
 * @dual_trigger         : Links needs to wait for two triggers prior to
 *                         applying the settings
 * @trigger_cnt          : trigger count value per device initiating the trigger
 * @eof_event_cnt        : Atomic variable to track the number of EOF requests
 * @skip_init_frame      : skip initial frames crm_wd_timer validation in the
 *                         case of long exposure use case
 * @last_sof_trigger_jiffies : Record the jiffies of last sof trigger jiffies
 * @wq_congestion        : Indicates if WQ congestion is detected or not
 * @try_for_internal_recovery : If the link stalls try for RT internal recovery
 * @properties_mask      : Indicates if current link enables some special properties
 * @cont_empty_slots     : Continuous empty slots
 * @is_shdr              : flag to indicate auto shdr usecase without SFE
 * @wait_for_dual_trigger: Flag to indicate whether to wait for second epoch in dual trigger
 */
struct cam_req_mgr_core_link {
	int32_t                              link_hdl;
	int32_t                              num_devs;
	enum cam_pipeline_delay              max_delay;
	enum cam_pipeline_delay              min_delay;
	enum cam_modeswitch_delay            max_mswitch_delay;
	enum cam_modeswitch_delay            min_mswitch_delay;
	struct cam_req_mgr_core_workq       *workq;
	int32_t                              pd_mask;
	struct cam_req_mgr_connected_device *l_dev;
	struct cam_req_mgr_req_data          req;
	struct cam_req_mgr_timer            *watchdog;
	struct completion                    workq_comp;
	enum cam_req_mgr_link_state          state;
	void                                *parent;
	struct mutex                         lock;
	spinlock_t                           link_state_spin_lock;
	struct cam_req_mgr_core_link
			*sync_link[MAXIMUM_LINKS_PER_SESSION - 1];
	int32_t                              num_sync_links;
	bool                                 sync_link_sof_skip;
	uint32_t                             open_req_cnt;
	int64_t                              last_flush_id;
	atomic_t                             is_used;
	bool                                 is_master;
	bool                                 initial_skip;
	bool                                 in_msync_mode;
	int64_t                              initial_sync_req;
	uint32_t                             retry_cnt;
	bool                                 is_shutdown;
	uint64_t                             sof_timestamp;
	uint64_t                             prev_sof_timestamp;
	bool                                 dual_trigger;
	uint32_t trigger_cnt[CAM_REQ_MGR_MAX_TRIGGERS]
				[CAM_TRIGGER_MAX_POINTS + 1];
	atomic_t                             eof_event_cnt;
	bool                                 skip_init_frame;
	uint64_t                             last_sof_trigger_jiffies;
	bool                                 wq_congestion;
	bool                                 try_for_internal_recovery;
	bool                                 is_sending_req;
	uint32_t                             properties_mask;
	uint32_t                             cont_empty_slots;
	bool                                 is_shdr;
	bool                                 wait_for_dual_trigger;
};

/**
 * struct cam_req_mgr_core_session
 * - Session Properties
 * @session_hdl        : session identifier
 * @num_links          : num of active links for current session
 * - Links of this session
 * @links              : pointer to array of links within session
 * - Session private data
 * @entry              : pvt data - entry in the list of sessions
 * @lock               : pvt data - spin lock to guard session data
 * - Debug data
 * @force_err_recovery : For debugging, we can force bubble recovery
 *                       to be always ON or always OFF using debugfs.
 * @sync_mode          : Sync mode for this session links
 */
struct cam_req_mgr_core_session {
	int32_t                       session_hdl;
	uint32_t                      num_links;
	struct cam_req_mgr_core_link *links[MAXIMUM_LINKS_PER_SESSION];
	struct list_head              entry;
	struct mutex                  lock;
	int32_t                       force_err_recovery;
	int32_t                       sync_mode;
};

/**
 * struct cam_req_mgr_core_device
 * - Core camera request manager data struct
 * @session_head : list head holding sessions
 * @crm_lock     : mutex lock to protect session creation & destruction
 * @recovery_on_apply_fail : Recovery on apply failure using debugfs.
 */
struct cam_req_mgr_core_device {
	struct list_head             session_head;
	struct mutex                 crm_lock;
	bool                         recovery_on_apply_fail;
};

/**
 * struct cam_req_mgr_req_tbl
 * @id            : table indetifier
 * @pd            : pipeline delay of table
 * @dev_count     : num of devices having same pipeline delay
 * @dev_mask      : mask to track which devices are linked
 * @skip_traverse : to indicate how many traverses need to be dropped
 *                  by this table especially in the beginning or bubble recovery
 * @pd_delta      : differnce between this table's pipeline delay and next
 * @num_slots     : number of request slots present in the table
 * @slot          : array of slots tracking requests availability at devices
 */
struct cam_req_mgr_req_tbl_mini_dump {
	int32_t                     id;
	int32_t                     pd;
	int32_t                     dev_count;
	int32_t                     dev_mask;
	int32_t                     skip_traverse;
	int32_t                     pd_delta;
	int32_t                     num_slots;
	struct cam_req_mgr_tbl_slot slot[MAX_REQ_SLOTS];
};

/**
 * struct cam_req_mgr_req_data_mini_dump
 * @apply_data       : Holds information about request id for a request
 * @prev_apply_data  : Holds information about request id for a previous
 *                     applied request
 * @in_q             : Poiner to Input request queue
 * @l_tbl            : unique pd request tables.
 * @num_tbl          : how many unique pd value devices are present
 */
struct cam_req_mgr_req_data_mini_dump {
	struct cam_req_mgr_apply       apply_data[CAM_PIPELINE_DELAY_MAX];
	struct cam_req_mgr_apply       prev_apply_data[CAM_PIPELINE_DELAY_MAX];
	struct cam_req_mgr_req_queue   in_q;
	struct cam_req_mgr_req_tbl_mini_dump l_tbl[4];
	int32_t                              num_tbl;
};

/**
 * struct cam_req_mgr_core_link_mini_dump
 * @workq                : Work q information
 * @req                  : req data holder.
 * @initial_sync_req     : The initial req which is required to sync with the
 * @prev_sof_timestamp   : Previous SOF timestamp value
 * @last_flush_id        : Last request to flush
 * @is_used              : 1 if link is in use else 0
 * @eof_event_cnt        : Atomic variable to track the number of EOF requests
 * @pd_mask              : each set bit indicates the device with pd equal to
 * @num_sync_links       : num of links sync associated with this link
 * @open_req_cnt         : Counter to keep track of open requests that are yet
 *                         to be serviced in the kernel.
 * @link_hdl             : Link identifier
 * @num_devs             : num of connected devices to this link
 * @max_delay            : Max of pipeline delay of all connected devs
 *                          bit position is available.
 * @state                : link state machine
 * @dual_trigger         : Dual tirgger flag
 * @is_shutdown          : Is shutting down
 * @skip_init_frame      : skip initial frames crm_wd_timer validation in the
 * @is_master            : Based on pd among links, the link with the highest pd
 *                         is assigned as master
 * @initial_skip         : Flag to determine if slave has started streaming in
 *                         master-slave sync
 * @in_m_sync_mode       : M-sync  mode flag
 * @sync_link_sof_skip   : flag determines if a pkt is not available
 * @wq_congestion        : Indicates if WQ congestion is detected or not
 */
struct cam_req_mgr_core_link_mini_dump {
	struct cam_req_mgr_core_workq_mini_dump  workq;
	struct cam_req_mgr_req_data_mini_dump    req;
	int64_t                   initial_sync_req;
	uint32_t                  last_flush_id;
	uint32_t                  is_used;
	uint32_t                  retry_cnt;
	uint32_t                  eof_event_cnt;
	int32_t                   pd_mask;
	int32_t                   num_sync_links;
	int32_t                   open_req_cnt;
	int32_t                   link_hdl;
	int32_t                   num_devs;
	enum cam_pipeline_delay   max_delay;
	enum cam_req_mgr_link_state   state;
	bool                       dual_trigger;
	bool                       is_shutdown;
	bool                       skip_init_frame;
	bool                       is_master;
	bool                       initial_skip;
	bool                       in_msync_mode;
	bool                       sync_link_sof_skip;
	bool                       wq_congestion;
};

/**
 * struct cam_req_mgr_core_mini_dump
 * @link             : Array of dumped links
 * @num_link         : Number of links dumped
 */
struct cam_req_mgr_core_mini_dump {
	struct cam_req_mgr_core_link_mini_dump *link[MAXIMUM_LINKS_PER_SESSION];
	uint32_t num_link;
};

/**
 * cam_req_mgr_create_session()
 * @brief    : creates session
 * @ses_info : output param for session handle
 *
 * called as part of session creation.
 */
int cam_req_mgr_create_session(struct cam_req_mgr_session_info *ses_info);

/**
 * cam_req_mgr_destroy_session()
 * @brief    : destroy session
 * @ses_info : session handle info, input param
 * @is_shutdown: To indicate if devices on link need to be disconnected.
 *
 * Called as part of session destroy
 * return success/failure
 */
int cam_req_mgr_destroy_session(struct cam_req_mgr_session_info *ses_info,
	bool is_shutdown);

/**
 * cam_req_mgr_link()
 * @brief     : creates a link for a session
 * @link_info : handle and session info to create a link
 *
 * link is formed in a session for multiple devices. it creates
 * a unique link handle for the link and is specific to a
 * session. Returns link handle
 */
int cam_req_mgr_link(struct cam_req_mgr_ver_info *link_info);
int cam_req_mgr_link_v2(struct cam_req_mgr_ver_info *link_info);


/**
 * cam_req_mgr_unlink()
 * @brief       : destroy a link in a session
 * @unlink_info : session and link handle info
 *
 * link is destroyed in a session
 */
int cam_req_mgr_unlink(struct cam_req_mgr_unlink_info *unlink_info);

/**
 * cam_req_mgr_schedule_request()
 * @brief: Request is scheduled
 * @sched_req: request id, session and link id info, bubble recovery info
 */
int cam_req_mgr_schedule_request(struct cam_req_mgr_sched_request *sched_req);

/**
 * cam_req_mgr_schedule_request_v2()
 * @brief: Request is scheduled
 * @sched_req: request id, session, link id info, bubble recovery info and sync info
 */
int cam_req_mgr_schedule_request_v2(struct cam_req_mgr_sched_request_v2 *sched_req);

/**
 * cam_req_mgr_sync_config()
 * @brief: sync for links in a session
 * @sync_info: session, links info and master link info
 */
int cam_req_mgr_sync_config(struct cam_req_mgr_sync_mode *sync_info);

/**
 * cam_req_mgr_flush_requests()
 * @brief: flush all requests
 * @flush_info: requests related to link and session
 */
int cam_req_mgr_flush_requests(struct cam_req_mgr_flush_info *flush_info);

/**
 * cam_req_mgr_core_device_init()
 * @brief: initialize crm core
 */
int cam_req_mgr_core_device_init(void);

/**
 * cam_req_mgr_core_device_deinit()
 * @brief: cleanp crm core
 */
int cam_req_mgr_core_device_deinit(void);

/**
 * cam_req_mgr_handle_core_shutdown()
 * @brief: Handles camera close
 */
void cam_req_mgr_handle_core_shutdown(void);

/**
 * cam_req_mgr_link_control()
 * @brief:   Handles link control operations
 * @control: Link control command
 */
int cam_req_mgr_link_control(struct cam_req_mgr_link_control *control);

/**
 * cam_req_mgr_dump_request()
 * @brief:   Dumps the request information
 * @dump_req: Dump request
 */
int cam_req_mgr_dump_request(struct cam_dump_req_cmd *dump_req);

/**
 * cam_req_mgr_link_properties()
 * @brief:   Handles link properties
 * @properties: Link properties
 */
int cam_req_mgr_link_properties(struct cam_req_mgr_link_properties *properties);

#endif
