/* SPDX-License-Identifier: GPL-2.0-only */
/*
 * Copyright (c) 2021-2022, Qualcomm Innovation Center, Inc. All rights reserved.
 * Copyright (c) 2012-2021, The Linux Foundation. All rights reserved.
 */

#ifndef _DP_POWER_H_
#define _DP_POWER_H_

#include "dp_parser.h"
#include "dp_pll.h"
#include "sde_power_handle.h"

/**
 * sruct dp_power - DisplayPort's power related data
 *
 * @dp_phy_gdsc: GDSC regulator
 * @init: initializes the regulators/core clocks/GPIOs/pinctrl
 * @deinit: turns off the regulators/core clocks/GPIOs/pinctrl
 * @clk_enable: enable/disable the DP clocks
 * @clk_status: check for clock status
 * @set_pixel_clk_parent: set the parent of DP pixel clock
 * @park_clocks: park all clocks driven by PLL
 * @clk_get_rate: get the current rate for provided clk_name
 * @power_client_init: configures clocks and regulators
 * @power_client_deinit: frees clock and regulator resources
 * @power_mmrm_init: configures mmrm client registration
 */
struct dp_power {
	struct drm_device *drm_dev;
	struct sde_power_handle *phandle;
	struct regulator *dp_phy_gdsc;
	int (*init)(struct dp_power *power, bool flip);
	int (*deinit)(struct dp_power *power);
	int (*clk_enable)(struct dp_power *power, enum dp_pm_type pm_type,
				bool enable);
	bool (*clk_status)(struct dp_power *power, enum dp_pm_type pm_type);
	int (*set_pixel_clk_parent)(struct dp_power *power, u32 stream_id);
	int (*park_clocks)(struct dp_power *power);
	u64 (*clk_get_rate)(struct dp_power *power, char *clk_name);
	int (*power_client_init)(struct dp_power *power,
		struct sde_power_handle *phandle,
		struct drm_device *drm_dev);
	void (*power_client_deinit)(struct dp_power *power);
	int (*power_mmrm_init)(struct dp_power *power,
                struct sde_power_handle *phandle, void *dp,
		int (*dp_display_mmrm_callback)(struct mmrm_client_notifier_data *notifier_data));
};

/**
 * dp_power_get() - configure and get the DisplayPort power module data
 *
 * @parser: instance of parser module
 * @pll: instance of pll module
 * return: pointer to allocated power module data
 *
 * This API will configure the DisplayPort's power module and provides
 * methods to be called by the client to configure the power related
 * modueles.
 */
struct dp_power *dp_power_get(struct dp_parser *parser, struct dp_pll *pll);

/**
 * dp_power_put() - release the power related resources
 *
 * @power: pointer to the power module's data
 */
void dp_power_put(struct dp_power *power);

#if defined(CONFIG_SECDP)
enum dp_hpd_plug_orientation secdp_get_plug_orientation(struct dp_power *dp_power);

int  secdp_power_request_gpios(struct dp_power *dp_power);
void secdp_power_set_gpio(struct dp_power *dp_power, bool flip);
void secdp_power_unset_gpio(struct dp_power *dp_power);

#if defined(CONFIG_SECDP_FACTORY_DPSWITCH_TEST)
void secdp_config_gpios_factory(struct dp_power *dp_power, int aux_sel, bool out_en);
#endif

void secdp_redriver_onoff(struct dp_power *dp_power, bool enable, int lane);
void secdp_redriver_linkinfo(struct dp_power *dp_power, u32 rate, u8 v_level, u8 p_level);
#endif/*CONFIG_SECDP*/

#endif /* _DP_POWER_H_ */
